@@ -36,9 +36,11 @@ TEMP_MINIMUM = 5
 TEMP_MAXIMUM = 30
 TEMP_OFF = -20
 
-TIMEOUT = 5
+DAYS = ["Monday","Tuesday","Wednesday","Thursday","Friday","Saturday","Sunday"]
 
-__VERSION__ = "1.0.7.2"
+TIMEOUT = 15
+
+__VERSION__ = "1.0.10"
 
 """
 Exception Handlers
@@ -94,7 +96,9 @@ class WiserNoRoomsFound(Error):
 
 class wiserHub:
     def __init__(self, hubIP, secret):
-        _LOGGER.info("WiserHub API Initialised : Version {}".format(__VERSION__))
+        _LOGGER.info(
+            "WiserHub API Initialised : Version {}".format(__VERSION__)
+        )
         self.wiserHubData = None
         self.wiserNetworkData = None
         self.hubIP = hubIP
@@ -136,6 +140,30 @@ class wiserHub:
         else:
             return True
 
+    def __getNextDay(self, day):
+        """
+        Finds the next day of week
+        param day: The day to convert
+        return: string
+        """
+        i = DAYS.index(day)
+        i += 1
+        if i > 6:
+            i = 0
+        return DAYS[i]
+
+    def __getPrevDay(self, day):
+        """
+        Finds the previous day of week
+        param day: The day to convert
+        return: string
+        """
+        i = DAYS.index(day)
+        i -= 1
+        if i < 0:
+            i = 6
+        return DAYS[i]
+
     def checkHubData(self):
         """
         Method checks the hub data object is populated, if it is not then it
@@ -146,7 +174,9 @@ class wiserHub:
         if self.wiserHubData is None:
             self.refreshData()
         if self.wiserHubData is None:
-            raise WiserHubDataNull("Hub data null even after refresh, aborting request")
+            raise WiserHubDataNull(
+                "Hub data null even after refresh, aborting request"
+            )
         # Otherwise continue
 
     def refreshData(self):
@@ -158,13 +188,17 @@ class wiserHub:
         _LOGGER.info("Updating Wiser Hub Data")
         try:
             resp = requests.get(
-                WISERHUBURL.format(self.hubIP), headers=self.headers, timeout=TIMEOUT
+                WISERHUBURL.format(self.hubIP),
+                headers=self.headers,
+                timeout=TIMEOUT,
             )
 
             resp.raise_for_status()
             self.wiserHubData = resp.json()
 
-            _LOGGER.debug("Wiser Hub Data received {} ".format(self.wiserHubData))
+            _LOGGER.debug(
+                "Wiser Hub Data received {} ".format(self.wiserHubData)
+            )
             if self.getRooms() is not None:
                 for room in self.getRooms():
                     roomStatId = room.get("RoomStatId")
@@ -193,7 +227,8 @@ class wiserHub:
             else:
                 _LOGGER.warning("Wiser found no rooms")
 
-            # The Wiser Heat Hub can return invalid JSON, so remove all non-printable characters before trying to parse JSON
+            # The Wiser Heat Hub can return invalid JSON, so remove all
+            # non-printable characters before trying to parse JSON
             responseContent = requests.get(
                 WISERNETWORKURL.format(self.hubIP),
                 headers=self.headers,
@@ -203,14 +238,16 @@ class wiserHub:
             self.wiserNetworkData = json.loads(responseContent)
 
         except requests.Timeout:
-            _LOGGER.debug("Connection timed out trying to update from Wiser Hub")
+            _LOGGER.debug(
+                "Connection timed out trying to update from Wiser Hub"
+            )
             raise WiserHubTimeoutException("The connection timed out.")
-        except requests.HTTPError:
-            if self.wiserHubData.status_code == 401:
+        except requests.HTTPError as ex:
+            if ex.response.status_code == 401:
                 raise WiserHubAuthenticationException(
                     "Authentication error.  Check secret key."
                 )
-            elif self.wiserHubData.status_code == 404:
+            elif ex.response.status_code == 404:
                 raise WiserRESTException("Not Found.")
             else:
                 raise WiserRESTException("Unknown Error.")
@@ -331,7 +368,9 @@ class wiserHub:
         """
         self.checkHubData()
         _LOGGER.debug(
-            " getDeviceRoom called, valve2roomMap is {} ".format(self.device2roomMap)
+            " getDeviceRoom called, valve2roomMap is {} ".format(
+                self.device2roomMap
+            )
         )
         if not self.device2roomMap:
             self.refreshData()
@@ -379,15 +418,20 @@ class wiserHub:
         DHWOffTemp = -200
 
         modeMapping = {
-            "on": {"RequestOverride": {"Type": "Manual", "SetPoint": DHWOnTemp}},
-            "off": {"RequestOverride": {"Type": "Manual", "SetPoint": DHWOffTemp}},
+            "on": {
+                "RequestOverride": {"Type": "Manual", "SetPoint": DHWOnTemp}
+            },
+            "off": {
+                "RequestOverride": {"Type": "Manual", "SetPoint": DHWOffTemp}
+            },
             "auto": {"RequestOverride": {"Type": "None", "Mode": "Auto"}},
         }
 
         _mode = mode.lower()
         if _mode not in ["on", "off", "auto"]:
             raise ValueError(
-                "Hot Water can be either 'on', 'off' or 'auto' - not '%s'" % _mode
+                "Hot Water can be either 'on', 'off' or 'auto' - not '%s'"
+                % _mode
             )
 
         # Obtain our DHW control ID
@@ -397,13 +441,17 @@ class wiserHub:
 
         _url = WISERHUBURL.format(self.hubIP) + "/HotWater/{}/".format(DHWId)
         _LOGGER.debug(
-            "Sending Patch Data: {}, to URL [{}]".format(modeMapping.get(_mode), _url)
+            "Sending Patch Data: {}, to URL [{}]".format(
+                modeMapping.get(_mode), _url
+            )
         )
         response = requests.patch(
             url=_url, headers=self.headers, json=modeMapping.get(_mode)
         )
         if response.status_code != 200:
-            _LOGGER.debug("Set DHW Response code = {}".format(response.status_code))
+            _LOGGER.debug(
+                "Set DHW Response code = {}".format(response.status_code)
+            )
             raise WiserRESTException(
                 "Error setting hot water mode to {}, error {} {}".format(
                     _mode, response.status_code, response.text
@@ -431,7 +479,9 @@ class wiserHub:
         )
         if response.status_code != 200:
             _LOGGER.debug(
-                "Set {} Response code = {}".format(switch, response.status_code)
+                "Set {} Response code = {}".format(
+                    switch, response.status_code
+                )
             )
             raise WiserRESTException(
                 "Error setting {} , error {} {}".format(
@@ -462,26 +512,138 @@ class wiserHub:
             "getRoomStatData for deviceID {} not found due".format(deviceId)
         )
 
+    def getSchedule(self, scheduleId):
+        """
+        Get Schedule by Schedule ID
+        
+        param scheduleId:
+        return: json data
+        """
+        self.checkHubData()
+        for schedule in self.wiserHubData.get("Schedule"):
+                if schedule.get("id") == scheduleId:
+                    return schedule
+        
+        raise WiserNotFound(
+            "Schedule with id {} not found ".format(scheduleId)
+        )
+
+    def setSchedule(self, scheduleId, scheduleData: dict):
+        """
+        Set Schedule by Schedule ID
+        
+        param scheduleId:
+        param scheduleData:
+        return: boolean
+        """
+        self.checkHubData()
+
+        for schedule in self.wiserHubData.get("Schedule"):
+                if schedule.get("id") == scheduleId:
+                    patchData = scheduleData
+                    response = requests.patch(
+                        url=WISERSCHEDULEURL.format(self.hubIP, scheduleId),
+                        headers=self.headers,
+                        json=patchData,
+                        timeout=TIMEOUT,
+                    )
+
+                    if response.status_code != 200:
+                        _LOGGER.debug(
+                            "Set Schedule Response code = {}".format(
+                                response.status_code
+                            )
+                        )
+                        raise WiserRESTException(
+                            "Error setting schedule for id {} , error {} {}".format(
+                                scheduleId, response.status_code, response.text
+                            )
+                        )
+                    else:
+                        return True
+        raise WiserNotFound("No schedule found that matches id")
+
+    def setScheduleFromFile(self, scheduleId, scheduleFile: str):
+        """
+        Sets Schedule from File
+
+        param scheduleId:
+        param scheduleData: json data for schedule
+        return: boolen
+        """
+        for schedule in self.wiserHubData.get("Schedule"):
+                if schedule.get("id") == scheduleId:
+                    if os.path.exists(scheduleFile):
+                        try:
+                            with open(scheduleFile, "r") as f:
+                                scheduleData = json.load(f)
+                        except:
+                            raise Exception(
+                                "Error reading file {}".format(scheduleFile)
+                            )
+
+                        patchData = scheduleData
+                        response = requests.patch(
+                            url=WISERSCHEDULEURL.format(self.hubIP, scheduleId),
+                            headers=self.headers,
+                            json=patchData,
+                            timeout=TIMEOUT,
+                        )
+
+                        if response.status_code != 200:
+                            _LOGGER.debug(
+                                "Set Schedule Response code = {}".format(
+                                    response.status_code
+                                )
+                            )
+                            raise WiserRESTException(
+                                "Error setting schedule {} , error {} {}".format(
+                                    scheduleId, response.status_code, response.text
+                                )
+                            )
+                        else:
+                            return True
+                    else:
+                        raise FileNotFoundError(
+                            "Schedule file, {}, not found.".format(
+                                os.path.abspath(scheduleFile)
+                            )
+                        )
+        raise WiserNotFound("No schedule found that matches Id")
+
+    def copySchedule(self, fromScheduleId, toScheduleId):
+        """
+        Copies schedule from one id to another
+
+        param fromScheduleId:
+        param toScheduleId:
+        return:
+        """
+        scheduleData = self.getSchedule(fromScheduleId)
+
+        if scheduleData is not None:
+            self.setSchedule(toScheduleId, scheduleData)
+        else:
+            raise WiserNotFound(
+                "Error copying schedule.  One of the schedule Ids is not valid"
+            )
+
     def getRoomSchedule(self, roomId):
         """
         Gets Room Schedule Data
-        
         param roomId:
         return: json data
         """
         self.checkHubData()
+        try:
+            scheduleId = self.getRoom(roomId).get("ScheduleId",0)
+            if scheduleId > 0:
+                return self.getSchedule(scheduleId)
+        except:
+            raise WiserNotFound(
+                "Schedule for room {} not found".format(roomId)
+            )
 
-        if self.getRoom(roomId) is None:
-            raise WiserNotFound("getRoomSchedule for room {} not found ".format(roomId))
-
-        scheduleId = self.getRoom(roomId).get("ScheduleId")
-        if scheduleId is not None:
-            for schedule in self.wiserHubData.get("Schedule"):
-                if schedule.get("id") == scheduleId:
-                    return schedule
-            raise WiserNotFound("getRoomSchedule for room {} not found ".format(roomId))
-        else:
-            raise WiserNotFound("getRoomSchedule for room {} not found ".format(roomId))
 
     def setRoomSchedule(self, roomId, scheduleData: dict):
         """
@@ -491,29 +653,71 @@ class wiserHub:
         param scheduleData: json data for schedule
         return:
         """
-        scheduleId = self.getRoom(roomId).get("ScheduleId")
-
-        if scheduleId is not None:
-            patchData = scheduleData
-            response = requests.patch(
-                url=WISERSCHEDULEURL.format(self.hubIP, scheduleId),
-                headers=self.headers,
-                json=patchData,
-                timeout=TIMEOUT,
-            )
-
-            if response.status_code != 200:
-                _LOGGER.debug(
-                    "Set Schedule Response code = {}".format(response.status_code)
-                )
-                raise WiserRESTException(
-                    "Error setting schedule for room {} , error {} {}".format(
-                        roomId, response.status_code, response.text
-                    )
-                )
+        scheduleId = self.getRoom(roomId).get("ScheduleId",0)
+        if scheduleId > 0:
+            self.setSchedule(scheduleId, scheduleData)
         else:
             raise WiserNotFound("No schedule found that matches roomId")
 
+    def setRoomScheduleAdvance(self, roomId):
+        """
+        Emulates the "Advance" button on a central heating controller
+	    Advances the temperature in a room to the next set point
+        param roomId: The id number of the room
+        """
+        name = self.getRoom(roomId).get("Name")
+        roomschedule = self.getRoomSchedule(roomId)
+        local_date_time = self.getSystem().get("LocalDateAndTime")
+        current_day = local_date_time.get("Day")
+        current_time = local_date_time.get("Time")
+        _LOGGER.info("{} {}".format(current_day,current_time))
+        set_points = roomschedule.get(current_day).get("SetPoints")
+        _LOGGER.info("{}".format(set_points))
+        for setpoint in set_points:
+            _LOGGER.debug("Set point time: {} temp: {}".format(setpoint.get("Time"), setpoint.get("DegreesC")))
+            if setpoint.get("Time") > current_time:
+                new_temp = round(setpoint.get("DegreesC") / 10, 1)
+                _LOGGER.info("Using set point at {}".format(setpoint.get("Time")))			
+                break
+        if set_points[-1].get("Time") <= current_time: # set point is next day
+            set_points = roomschedule.get(self.__getNextDay(current_day)).get("SetPoints")
+            _LOGGER.info(
+               "Using set point on {} at {}".format(self.__getNextDay(current_day), set_points[0].get("Time"))
+            )
+            new_temp = round(set_points[0].get("DegreesC") /10, 1) # use first setpoint temp of next day
+					
+        _LOGGER.info("Advance {} to {} DegreesC".format(name, new_temp))
+        self.setRoomTemperature(roomId, new_temp)
+
+    def setRoomScheduleAdvanceUndo(self, roomId):
+        """
+        Undoes any Advance of the temperature in the room, reverts it back to the scheduled set temperature
+        param roomId: The id number of the room
+        """
+        name = self.getRoom(roomId).get("Name")
+        roomschedule = self.getRoomSchedule(roomId)
+        local_date_time = self.getSystem().get("LocalDateAndTime")
+        current_day = local_date_time.get("Day")
+        current_time = local_date_time.get("Time")
+        _LOGGER.info("{} {}".format(current_day,current_time))
+        set_points = roomschedule.get(current_day).get("SetPoints")
+        _LOGGER.info("{}".format(set_points))
+        for setpoint in reversed(set_points):
+            _LOGGER.debug("{} {}".format(setpoint.get("Time"), setpoint.get("DegreesC")))
+            if setpoint.get("Time") <= current_time:
+                new_temp = round(setpoint.get("DegreesC") / 10, 1)
+                _LOGGER.info("Using set point at {}".format(setpoint.get("Time")))
+                break
+        if set_points[0].get("Time") > current_time: # set point is prev day
+            set_points = roomschedule.get(self.__getPrevDay(current_day)).get("SetPoints")
+            _LOGGER.info(
+                "Using set point on {} at {}".format(self.__getPrevDay(current_day), set_points[-1].get("Time"))
+            )
+            new_temp = round(set_points[-1].get("DegreesC") /10, 1) # use last setpoint temp of prev day
+
+        _LOGGER.info("AdvanceUndo {} to {} DegreesC".format(name, new_temp))
+        self.setRoomTemperature(roomId, new_temp)
+
     def setRoomScheduleFromFile(self, roomId, scheduleFile: str):
         """
         Sets Room Schedule
@@ -522,39 +726,10 @@ class wiserHub:
         param scheduleData: json data for schedule
         return:
         """
-        scheduleId = self.getRoom(roomId).get("ScheduleId")
+        scheduleId = self.getRoom(roomId).get("ScheduleId",0)
 
-        if scheduleId is not None:
-            if os.path.exists(scheduleFile):
-                try:
-                    with open(scheduleFile, "r") as f:
-                        scheduleData = json.load(f)
-                except:
-                    raise Exception("Error reading file {}".format(scheduleFile))
-
-                patchData = scheduleData
-                response = requests.patch(
-                    url=WISERSCHEDULEURL.format(self.hubIP, scheduleId),
-                    headers=self.headers,
-                    json=patchData,
-                    timeout=TIMEOUT,
-                )
-
-                if response.status_code != 200:
-                    _LOGGER.debug(
-                        "Set Schedule Response code = {}".format(response.status_code)
-                    )
-                    raise WiserRESTException(
-                        "Error setting schedule for room {} , error {} {}".format(
-                            roomId, response.status_code, response.text
-                        )
-                    )
-            else:
-                raise FileNotFoundError(
-                    "Schedule file, {}, not found.".format(
-                        os.path.abspath(scheduleFile)
-                    )
-                )
+        if scheduleId > 0:
+            self.setScheduleFromFile(scheduleId, scheduleFile)
         else:
             raise WiserNotFound("No schedule found that matches roomId")
 
@@ -566,19 +741,15 @@ class wiserHub:
         param toRoomId:
         return: boolean
         """
-        scheduleData = self.getRoomSchedule(fromRoomId)
-
-        print(json.dumps(scheduleData))
-
-        print("TYPE:{}".format(type(scheduleData)))
-
-        if scheduleData is not None:
-            self.setRoomSchedule(toRoomId, scheduleData)
+        fromScheduleId = self.getRoom(fromRoomId).get("ScheduleId",0)
+        toScheduleId = self.getRoom(toRoomId).get("ScheduleId",0)
+        if fromScheduleId > 0 and toScheduleId > 0:
+            self.copySchedule(fromScheduleId, toScheduleId)
         else:
             raise WiserNotFound(
                 "Error copying schedule.  One of the room Ids is not valid"
             )
-
+ 
     def setHomeAwayMode(self, mode, temperature=10):
         """
         Sets default Home or Away mode, optionally allows you to set a temperature for away mode
@@ -589,14 +760,18 @@ class wiserHub:
 
         return:
         """
-        _LOGGER.info("Setting Home/Away mode to : {} {} C".format(mode, temperature))
+        _LOGGER.info(
+            "Setting Home/Away mode to : {} {} C".format(mode, temperature)
+        )
 
         if mode not in ["HOME", "AWAY"]:
             raise ValueError("setAwayHome can only be HOME or AWAY")
 
         if mode == "AWAY":
             if temperature is None:
-                raise ValueError("setAwayHome set to AWAY but not temperature set")
+                raise ValueError(
+                    "setAwayHome set to AWAY but not temperature set"
+                )
             if not (self.__checkTempRange(temperature)):
                 raise ValueError(
                     "setAwayHome temperature can only be between {} and {} or {}(Off)".format(
@@ -606,7 +781,10 @@ class wiserHub:
         _LOGGER.info("Setting Home/Away : {}".format(mode))
 
         if mode == "AWAY":
-            patchData = {"type": 2, "setPoint": self.__toWiserTemp(temperature)}
+            patchData = {
+                "type": 2,
+                "setPoint": self.__toWiserTemp(temperature),
+            }
         else:
             patchData = {"type": 0, "setPoint": 0}
         _LOGGER.debug("patchdata {} ".format(patchData))
@@ -632,7 +810,9 @@ class wiserHub:
         param roomId:  The Room ID
         param temperature:  The temperature in celcius from 5 to 30, -20 for Off
         """
-        _LOGGER.info("Set Room {} Temperature to = {} ".format(roomId, temperature))
+        _LOGGER.info(
+            "Set Room {} Temperature to = {} ".format(roomId, temperature)
+        )
         if not (self.__checkTempRange(temperature)):
             raise ValueError(
                 "SetRoomTemperature : value of temperature must be between {} and {} OR {} (off)".format(
@@ -661,7 +841,9 @@ class wiserHub:
                 "Error setting temperature, error {} ".format(response.text)
             )
         _LOGGER.debug(
-            "Set room Temp, error {} ({})".format(response.status_code, response.text)
+            "Set room Temp, error {} ({})".format(
+                response.status_code, response.text
+            )
         )
 
     # Set Room Mode (Manual, Boost,Off or Auto )
@@ -706,14 +888,22 @@ class wiserHub:
                 }
             }
         elif mode.lower() == "manual":
-            # When setting to manual , set the temp to the current scheduled temp
-            setTemp = self.__fromWiserTemp(
-                self.getRoom(roomId).get("ScheduledSetPoint")
-            )
+            # Set to manual mode first if currently auto
+            if self.getRoom(roomId).get("Mode") == "Auto":
+                self.setRoomMode(roomId, "auto_to_manual")
+                # If already manual (ie off), set the temp to the current scheduled temp
+                setTemp = self.__fromWiserTemp(
+                    self.getRoom(roomId).get("CurrentSetPoint")
+                )
+            else:
+                # If already manual (ie off), set the temp to the current scheduled temp
+                setTemp = self.__fromWiserTemp(
+                    self.getRoom(roomId).get("ScheduledSetPoint")
+                )
+
             # If current scheduled temp is less than 5C then set to min temp
             setTemp = setTemp if setTemp >= TEMP_MINIMUM else TEMP_MINIMUM
             patchData = {
-                "Mode": "Manual",
                 "RequestOverride": {
                     "Type": "Manual",
                     "SetPoint": self.__toWiserTemp(setTemp),
@@ -721,13 +911,18 @@ class wiserHub:
             }
         # Implement trv off as per https://github.com/asantaga/wiserheatingapi/issues/3
         elif mode.lower() == "off":
+            # Set to manual mode first if currently auto
+            if self.getRoom(roomId).get("Mode") == "Auto":
+                self.setRoomMode(roomId, "auto_to_manual")
+
             patchData = {
-                "Mode": "Manual",
                 "RequestOverride": {
                     "Type": "Manual",
                     "SetPoint": self.__toWiserTemp(TEMP_OFF),
                 },
             }
+        elif mode.lower() == "auto_to_manual":
+            patchData = {"Mode": "Manual"}
         else:
             raise ValueError(
                 "Error setting setting room mode, received  {} but should be auto,boost,off or manual ".format(
@@ -754,9 +949,13 @@ class wiserHub:
             )
             if response.status_code != 200:
                 _LOGGER.error(
-                    "Cancelling boost resulted in {}".format(response.status_code)
+                    "Cancelling boost resulted in {}".format(
+                        response.status_code
+                    )
+                )
+                raise WiserRESTException(
+                    "Error cancelling boost {} ".format(mode)
                 )
-                raise WiserRESTException("Error cancelling boost {} ".format(mode))
 
         # Set new mode
         response = requests.patch(
@@ -772,10 +971,14 @@ class wiserHub:
                 )
             )
             raise WiserRESTException(
-                "Error setting mode to {}, error {} ".format(mode, response.text)
+                "Error setting mode to {}, error {} ".format(
+                    mode, response.text
+                )
             )
         _LOGGER.debug(
-            "Set room mode, error {} ({})".format(response.status_code, response.text)
+            "Set room mode, error {} ({})".format(
+                response.status_code, response.text
+            )
         )
 
     def getSmartPlugs(self):
@@ -811,7 +1014,9 @@ class wiserHub:
     def setSmartPlugState(self, smartPlugId, smartPlugState):
         if smartPlugState is None:
             _LOGGER.error("SmartPlug State is None, must be either On or Off")
-            raise ValueError("SmartPlug State is None, must be either On or Off")
+            raise ValueError(
+                "SmartPlug State is None, must be either On or Off"
+            )
         if smartPlugState.title() not in ["On", "Off"]:
             _LOGGER.error("SmartPlug State must be either On or Off")
             raise ValueError("SmartPlug State must be either On or Off")
@@ -819,7 +1024,9 @@ class wiserHub:
         url = WISERSMARTPLUGURL.format(self.hubIP, smartPlugId)
         patchData = {"RequestOutput": smartPlugState.title()}
 
-        _LOGGER.debug("Setting smartplug status patchdata {} ".format(patchData))
+        _LOGGER.debug(
+            "Setting smartplug status patchdata {} ".format(patchData)
+        )
         response = requests.patch(
             url=url.format(self.hubIP),
             headers=self.headers,
@@ -862,7 +1069,9 @@ class wiserHub:
         url = WISERSMARTPLUGURL.format(self.hubIP, smartPlugId)
         patchData = {"Mode": smartPlugMode.title()}
 
-        _LOGGER.debug("Setting smartplug status patchdata {} ".format(patchData))
+        _LOGGER.debug(
+            "Setting smartplug status patchdata {} ".format(patchData)
+        )
         response = requests.patch(
             url=url.format(self.hubIP),
             headers=self.headers,
@@ -886,3 +1095,5 @@ class wiserHub:
                         response.status_code, response.text
                     )
                 )
+
+	
